#!/usr/bin/env python

#    Copyright (c) 2014-2017 Max Beloborodko.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__author__ = 'f1ashhimself@gmail.com'


from platform import system

uisoup = None


class TooSaltyUISoupException(Exception):
    pass

if system() == 'Windows':
    from uisoup.win_soup import WinSoup
    uisoup = WinSoup()
elif system() == 'Darwin':
    from uisoup.mac_soup import MacSoup
    uisoup = MacSoup()
else:
    raise TooSaltyUISoupException('We are sorry but we don\'t have UISoup '
                                  'implementation for "%s" OS.' % system())
